#include <hammer/hammer.h>
#include <hammer/glue.h>
#include <stdlib.h>	/* malloc, free */
#include <string.h>	/* memcpy */
#include "lzw.h"


struct context {
	/*
	 * Storing byte sequences represented by each LZW code.
	 *
	 * Codes 0-255 are predefined representing literals.
	 * Codes 256 and 267 are the special clear and eod (end of data) codes.
	 * Codes >257 are dynamically defined by the input.
	 *
	 * Each dynamically defined code is an extension of a previously
	 * defined code. We therefore need only store the code being extended
	 * and the byte being added.
	 *
	 * Thus the entries of this array form linked lists. To terminate the
	 * lists while making our memory allocation easy, we store the length
	 * of each code's output sequence.
	 *
	 * Finally, we redundantly store the first byte of the sequence so we
	 * don't have to walk the list during updates (see act_output).
	 */
	struct {
		size_t len;	/* length of the sequence */
		int prefix;	/* code representing the seq's prefix (len-1) */
		uint8_t last;	/* final byte of the sequence */
		uint8_t first;	/* first byte of the sequence */
	} table[4097];		/* 4096 codes + one dummy (see act_output) */

	/*
	 * The next code to be assigned, i.e. the current size of the table.
	 */
	int next;

	/*
	 * earlychange = 1 means the bit size is increased "one code early".
	 * earlychange = 0 is "code length increases shall be postponed as
	 * long as possible".
	 */
	int earlychange;
};


/*
 * Helpers for working with the table:
 */

static void
lzw_clear_table(struct context *ctx)
{
	ctx->next = 258;
}

/*
 * Update the dictionary with a new entry that extends the given code by one
 * byte to be filled in later.
 */
static void
lzw_table_extend(struct context *ctx, int code)
{
	ctx->table[ctx->next].prefix = code;
	ctx->table[ctx->next].first = ctx->table[code].first;
	ctx->table[ctx->next].len = ctx->table[code].len + 1;
	ctx->table[ctx->next].last = 0xFF;
	ctx->next++;
}

/*
 * Assemble the output sequence represented by the given code word.
 * The given buffer must have the appropriate size.
 * Returns the number of bytes written.
 */
static size_t
lzw_code_string(struct context *ctx, int code, uint8_t *buf)
{
	size_t i, n;

	/* traverse the list, filling buf from last to first byte */
	n = ctx->table[code].len;
	for (i = 0; i < n; i++) {
		buf[n - 1 - i] = ctx->table[code].last;
		code = ctx->table[code].prefix;
	}
	assert(code == -1);	/* reached the end */

	return n;
}


/*
 * Global variables:
 */

HParser *p_lzwdata;
static struct context *context;


/*
 * Semantic actions and validations:
 */

static HParsedToken *
act_clear(const HParseResult *p, void *u)
{
	struct context *ctx = u;
	lzw_clear_table(ctx);
	return NULL;
}

static bool
validate_clear(HParseResult *p, void *u)
{
	uint64_t code = H_CAST_UINT(p->ast);
	return (code == 256);
}

static bool
validate_eod(HParseResult *p, void *u)
{
	uint64_t code = H_CAST_UINT(p->ast);
	return (code == 257);
}

static bool
validate_output(HParseResult *p, void *u)
{
	uint64_t code = H_CAST_UINT(p->ast);
	struct context *ctx = u;

	return (code != 256 && code != 257 && code < ctx->next);
}

static HParsedToken *
act_output(const HParseResult *p, void *u)
{
	uint64_t code = H_CAST_UINT(p->ast);
	struct context *ctx = u;

	assert(ctx->next >= 258);
	assert(ctx->next <= 4097);
	assert(code < ctx->next);
	assert(code != 256);
	assert(code != 257);

	/* Fill in the missing last byte of the last assigned code, if any. */
	if (ctx->next > 258)
		ctx->table[ctx->next - 1].last = ctx->table[code].first;

	/*
	 * Update the dictionary with a new entry that is missing the last
	 * byte which we will only learn when we process the next code.
	 *
	 * Note that the value 4097 is intentional here. Rather than going
	 * through the effort of ensuring that the last code is only updated
	 * once, we simply assign one more code as a dummy.
	 */
	if (ctx->next < 4097)
		lzw_table_extend(ctx, code);

	/* Pass the code through. Output is generated in act_lzwblock below. */
	return (HParsedToken *)p->ast;	// XXX casting away the const OK?
}

/*
 * Assemble the string represented by a block of code words under a given
 * table. The incoming HParsedToken is a sequence of code words (TT_UINT).
 */
static HParsedToken *
act_lzwblock(const HParseResult *p, void *u)
{
	HCountedArray *seq = H_CAST_SEQ(p->ast);
	struct context *ctx = u;
	uint8_t *buf, *cur;
	size_t sz, i;
	int code;

	/* determine total output size, alloc buffer */
	sz = 0;
	for (i = 0; i < seq->used; i++) {
		code = (int) H_CAST_UINT(seq->elements[i]);
		sz += ctx->table[code].len;
	}
	buf = h_arena_malloc(p->arena, sz);

	/* go through sequence, merge output bytes into buf */
	cur = buf;
	for (i = 0; i < seq->used; i++) {
		code = (int) H_CAST_UINT(seq->elements[i]);
		cur += lzw_code_string(ctx, code, cur);
	}
	assert(cur == buf + sz);

	return H_MAKE_BYTES(buf, sz);
}

/*
 * Concatenate blocks to form the final output string.
 * The incoming HParsedToken is a sequence of HBytes.
 */
static HParsedToken *
act_lzwdata(const HParseResult *p, void *u)
{
	HCountedArray *seq = H_CAST_SEQ(p->ast);
	HBytes bs;
	uint8_t *buf, *cur;
	size_t sz, i;

	/* fast path: single element? nothing to do */
	if (seq->used == 1)
		return seq->elements[0];

	/* determine total output size, alloc buffer */
	sz = 0;
	for (i = 0; i < seq->used; i++)
		sz += H_CAST_BYTES(seq->elements[i]).len;
	buf = h_arena_malloc(p->arena, sz);

	/* go through sequence, copying bytes into buf */
	cur = buf;
	for (i = 0; i < seq->used; i++) {
		bs = H_CAST_BYTES(seq->elements[i]);
		memcpy(cur, bs.token, bs.len);
		cur += bs.len;
	}
	assert(cur == buf + sz);

	return H_MAKE_BYTES(buf, sz);
}

/*
 * Continuation for h_bind() in the 'codeword' grammar rule. It inspects the
 * lzw context passed as 'env' and returns the parser of the correct size.
 *
 * NB: We create the returned parsers statically in init_LZW_parser() to avoid
 * allocation during the parse.
 */
static HParser *p_code9, *p_code10, *p_code11, *p_code12;
static HParser *
kcodeword(HAllocator *mm__, const HParsedToken *x, void *env)
{
	struct context *ctx = env;

	if (ctx->next <= 512 - ctx->earlychange)
		return p_code9;
	else if (ctx->next <= 1024 - ctx->earlychange)
		return p_code10;
	else if (ctx->next <= 2048 - ctx->earlychange)
		return p_code11;
	else
		return p_code12;
}


/*
 * Exposed interface:
 */

void
init_LZW_parser()
{
	int i;

	/* initialize global context variable, incl. static table entries */
	context = malloc(sizeof *context);
	assert(context != NULL);
	for(i = 0; i < 256; i++)
	{
		context->table[i].len = 1;
		context->table[i].prefix = -1;	/* none */
		context->table[i].first = i;
		context->table[i].last = i;
	}
	init_LZW_context(1);

	/* static parsers for code words of all possible sizes */
	p_code9  = h_bits(9,  false);
	p_code10 = h_bits(10, false);
	p_code11 = h_bits(11, false);
	p_code12 = h_bits(12, false);

	/* kcodeword() selects the appropriate parser based on context */
	H_RULE   (codeword,	h_bind(h_epsilon_p(), kcodeword, context));

	H_VRULE  (eod,		codeword);
	H_AVDRULE(clear,	codeword, context);
	H_AVDRULE(output,	codeword, context);

	H_ADRULE(lzwblock,	h_right(clear, h_many(output)), context);
	H_ARULE (lzwdata,	h_left(h_many1(lzwblock), eod));
	    // XXX validate that the last byte is zero-padded?
	    // XXX require h_end_p()?

	p_lzwdata = lzwdata;
}

HParseResult *
parse_LZW_data(const uint8_t *input, size_t length)
{
	HParseResult *res = h_parse(p_lzwdata, input, length);
	return res;
}

void
init_LZW_context(int earlychange)
{
	lzw_clear_table(context);
	context->earlychange = earlychange;
}
